// Copyright (C) 1999-2002 Open Source Telecom Corporation.
//  
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software 
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
// 
// As a special exception to the GNU General Public License, permission is 
// granted for additional uses of the text contained in its release 
// of Common C++.
// 
// The exception is that, if you link the Common C++ library with other files
// to produce an executable, this does not by itself cause the
// resulting executable to be covered by the GNU General Public License.
// Your use of that executable is in no way restricted on account of
// linking the Common C++ library code into it.
// 
// This exception does not however invalidate any other reasons why
// the executable file might be covered by the GNU General Public License.
// 
// This exception applies only to the code released under the 
// name Common C++.  If you copy code from other releases into a copy of
// Common C++, as the General Public License permits, the exception does
// not apply to the code that you add in this way.  To avoid misleading
// anyone as to the status of such modified files, you must delete
// this exception notice from them.
// 
// If you write modifications of your own for Common C++, it is your choice
// whether to permit this exception to apply to your modifications.
// If you do not wish that, delete this exception notice.


#include <cc++/config.h>
#include <cc++/export.h>
#include <cc++/thread.h>
#include <cc++/misc.h>
#include <cc++/strchar.h>
#include <cc++/string.h>
#include "private.h"

#include <cstdarg>
#include <cstdio>

#ifdef	CCXX_NAMESPACES
namespace ost {
using std::streambuf;
using std::ofstream;
using std::ostream;
using std::endl;
using std::ios;
#endif

static Mutex mutex;
MemPager *String::pager = NULL;
char **String::idx = NULL;

const unsigned String::minsize = ((sizeof(char *) + (sizeof(unsigned) * 2) + 1));
const unsigned String::slotsize = 32;
const unsigned String::pagesize = 1024;
const unsigned String::slotlimit = 512;
const unsigned String::slots = ((slotlimit / slotsize) + 1);
const unsigned String::npos = (unsigned)(-1);

String::String()
{
	init();
}

String::String(unsigned count, const char chr)
{
	init();
	resize(count + 1);
	memset(getText(), chr, count);
	setLength(count);
}

#ifdef	HAVE_SNPRINTF
String::String(unsigned count, const char *format, ...)
{
        va_list args;
        va_start(args, format);

	init();
	resize(count);

        char *ptr = getText();
	vsnprintf(ptr, count, format, args);
        setLength(strlen(ptr));
	va_end(args);
}
#else
String::String(unsigned count, const char *str)
{
        init();
        resize(count);

        if(!str || !*str)
                return;

        set(str);
}
#endif

String::String(const char *str)
{
	init();
	set(str);
}

String::String(const String &str, unsigned start, unsigned count)
{
	init();
	char *ptr = str.getText();
	unsigned len = str.getLength();

	if(start >= len)
		return;

	ptr += start;
	len -= start;
	
	if(count >= len)
		count = len;

	set(ptr, count);
}

char String::at(int index) const
{
	if(index < 0)
		index = getLength() - index + 1;
	if(index > getLength() || index < 0)
		return 0;

	return (getText())[index];
}

void String::insert(unsigned start, const char *str, unsigned count)
{
	char *ptr = getText();
	unsigned len = getLength();
	unsigned size = getSize();

	if(!str)
		str = "";

	if(!count)
		count = strlen(str);

	if(!count || start > len)
		return;

	if(len + count >= size)
	{
		resize(len + count + 1);
		ptr = getText();
	}

	if(start == len)
	{
		memcpy(ptr + start, str, count);
		len += count;
		setLength(len);
		ptr[len] = 0;
		return;
	}

	memmove(ptr + start + count, ptr + start, len - start);
	memmove(ptr + start, str, count);
	len += count;
	setLength(len);
	ptr[len] = 0;
	return;
}	


void String::insert(unsigned start, const String &s)
{
	insert(start, s.getText(), s.getLength());
}

void String::replace(unsigned start, unsigned len, const char *text, unsigned count)
{
	erase(start, len);
	insert(start, text, count);
}

void String::replace(unsigned start, unsigned len, const String &s)
{
	erase(start, len);
	insert(start, s);
}

void String::erase(unsigned start, unsigned count)
{
	unsigned len = getLength();
	char *ptr = getText();
		
	if(start >= len)
		return;

	if(start + count >= len || count == npos || !count)
	{
		setLength(start);
		ptr[start] = 0;
		return;
	}		

	memcpy(ptr + start, ptr + start + count, len - start - count);
	len -= count;
	setLength(len);
	ptr[len] = 0;
}

void String::append(const char *str, unsigned offset, unsigned len)
{
	unsigned slen = getLength();
	char *ptr = getText();

	if(slen < offset)
	{
		append(str, len);
		return;
	}

	setLength(offset);
	ptr[offset] = 0;
	append(str, len);
}

void String::append(const char *str, unsigned len)
{
	if(!str)
		return;

	if(!len)
		len = strlen(str);

	if(!len)
		return;

	if(getLength() + len >= getSize())
		resize(getLength() + len + 1);
	memcpy(getText() + getLength(), str, len);
	len += getLength();
	setLength(len);
	getText()[len] = 0;
	return;
}

void String::add(char c)
{
	unsigned len = getLength();
	char *ptr;

	if(len + 1 >= getSize())
		resize(len + 2);
	
	ptr = getText();
	ptr[len++] = c;
	setLength(len);
	ptr[len] = 0;
}

void String::trim(unsigned count)
{
	unsigned len = getLength();
	char *ptr = getText();

	if(count >= len)
		count = len;

	len -= count;
	ptr[len] = 0;
	setLength(len);
}

String String::token(const char *delim, unsigned offset)
{
        char *ptr = getText();
        unsigned len = getLength();
        unsigned count = 0;
        String result;
	bool found = false;

	if(offset >= len)
		return result;

	len -= offset;
	ptr += offset;
                                                                        
        while(count < len)
        {
                if(strchr(delim, ptr[count]))
		{
			found = true;
                        break;
		}
                ++count;
        }

	if(!count && found)
		erase(offset, 1);
                                                                        
        if(!count)
		return result;

        result.set(ptr, count);                                                                        
	if(found)
		++count;
        erase(offset, count);
	return result;
}

void String::strip(const char *chars)
{
	unsigned len = strtrim(chars, getText(), getLength());
	if(!len)
	{
		setLength(len);
		return;
	}
	setLength(strchop(chars, getText(), len));
}

#ifdef	HAVE_SNPRINTF
const char *String::set(unsigned size, const char *format, ...)
{
        va_list args;
        va_start(args, format);

	if(size <= minsize)
		clear();

	if(size > getSize())
		resize(size);

        char *ptr = getText();
	vsnprintf(ptr, size, format, args);
        setLength(strlen(ptr));
	va_end(args);
	return ptr;
}

void String::append(unsigned size, const char *format, ...)
{
        va_list args;
        va_start(args, format);
	unsigned len = getLength();

	if(len + size <= minsize)
		clear();

	if(len + size > getSize())
		resize(size + size);

        char *ptr = getText() + len;
	vsnprintf(ptr, size, format, args);
        setLength(strlen(getText()));
	va_end(args);
}

#endif
const char *String::set(const char *str, unsigned len)
{
	if(!str)
	{
		clear();
		return str;
	}

	if(!len)
		len = strlen(str);

	// if we are making a short string, lets clear prior alloc

	if(len < minsize)
		clear();

	if(len >= getSize())
		resize(len + 1);
	memcpy(getText(), str, len);
	getText()[len] = 0;
	setLength(len);
	return str;
}

void String::set(const String &str)
{
	set(str.getText(), str.getLength());
}

void String::append(const String &str)
{
	append(str.getText(), str.getLength());
}

void String::copy(const String &original)
{
	char *ptr;
	content = original.content;

	if(!isBig())
		return;

	if(getLength() <= minsize)
	{
		content.ministring.length = getLength();
		memcpy(content.ministring.text, getText(), getLength());
		content.ministring.big = false;
		return;
	}

	ptr = getText();
	content.bigstring.size = setSize(content.bigstring.length);
	content.bigstring.text = getSpace(getSize());
	memcpy(content.bigstring.text, ptr, getLength());
}

String::~String()
{
	if(isBig())
		clear();
}

void String::init(void)
{
	content.ministring.big = false;
	content.ministring.length = 0;
	content.ministring.text[0] = 0;
}

unsigned String::search(const char *text, unsigned clen, unsigned index) const
{
	unsigned len = getLength();
	 
	if(!text)
		text = "";

	if(!clen)
		clen = strlen(text);

	while(clen + index <= len)
	{
		if(compare(text, clen, index) == 0)
			return index;
		++index;
	}
	return npos;
}

unsigned String::find(const char *text, unsigned index, unsigned len, unsigned count) const
{
	unsigned pos;

	if(!text)
		text = "";

	if(!len)
		len = strlen(text);

	while(count--)
	{
		pos = search(text, len, index);
		if(pos == npos)
			break;
		index = pos + 1;
	}
	return pos;
}

unsigned String::rfind(const char *text, unsigned index, unsigned len) const
{
	unsigned result = npos;

        if(!text)
                text = "";

        if(!len)
                len = strlen(text);

	for(;;)
	{
		index = search(text, len, index);
		if(index == npos)
			break;
		result = index++;
	}
	return result;
}

unsigned String::count(const char *text, unsigned index, unsigned len) const
{
	unsigned count = 0;
	if(!text)
		text = "";

	if(!len)
		len = strlen(text);

	for(;;)
	{
		index = search(text, len, index);
		if(index == npos)
			break;
		++count;
		++index;
	}
	return count;
}

unsigned String::count(const String &str, unsigned index) const
{
	return count(str.getText(), index, str.getLength());
}

unsigned String::find(const String &str, unsigned index, unsigned count) const
{
	return find(str.getText(), index, str.getLength(), count);
}

unsigned String::rfind(const String &str, unsigned index) const
{
	return rfind(str.getText(), index, str.getLength());
}

int String::compare(const char *text, unsigned len, unsigned index) const
{
	if(!text)
		text = "";

	if(index > getLength())
		return -1;

	if(!len)
		return strcmp(getText() + index, text);

	return strncmp(getText() + index, text, len);
}

bool String::isEmpty(void) const
{
	char *ptr = getText();

	if(!ptr || !*ptr)
		return true;

	return false;
}

void String::resize(unsigned size)
{
	unsigned length = getLength();
	char *ptr;

	if(length >= size)
		length = size - 1;

	++length;

	if(!isBig() && size <= minsize)
		return;

	if(!isBig())
	{
		ptr = getSpace(size);
		memcpy(ptr, content.ministring.text, length); 
		ptr[--length] = 0;
		content.ministring.big = true;
		content.bigstring.text = ptr;
		content.bigstring.length = length;
		setSize(size);
		return;
	}
	
	if(size <= minsize && getSize() > slotlimit)
	{
		ptr = getText();
		memcpy(content.ministring.text, ptr, length);
		content.ministring.text[--length] = 0;
		content.ministring.big = false;
		content.ministring.length = length;
		delete[] ptr;
		return;
	}

	ptr = getSpace(size);
	memcpy(ptr, getText(), length);
	ptr[--length] = false;
	clear();
	setSize(size);
	content.bigstring.length = length;
	content.bigstring.text = ptr;
	content.ministring.big = true;
}

void String::clear(void)
{
	char **next;
	unsigned slot;

	if(!isBig())
		goto end;

	if(!content.bigstring.text)
		goto end; 

	if(getSize() > slotlimit)
	{
		delete[] content.bigstring.text;
		goto end;
	}

	slot = getSize() / slotsize;
	next = (char **)content.bigstring.text;
	mutex.enterMutex();
	*next = idx[slot];
	idx[slot] = content.bigstring.text;
	mutex.leaveMutex();	
	
end:
	init();
	return;
}	

char *String::getSpace(unsigned size)
{
	unsigned slot;
	char *text, **next;

	if(size > slotlimit)
		return new char[size];

	slot = size / slotsize;
	mutex.enterMutex();
	if(!pager)
	{
		pager = new MemPager(pagesize);
		idx = (char **)pager->alloc(sizeof(char *) * slots);
		memset(idx, 0, sizeof(char *) * slots);
	}
	text = idx[slot];
	if(text)
	{
		next = (char **)text;
		idx[slot] = *next;
	}
	else
		text = (char *)pager->alloc(++slot * slotsize);
	mutex.leaveMutex();
	return text;
}

const char *String::getIndex(unsigned index) const
{
	const char *dp = getText();

	if(index > getLength())
		return NULL;

	return (const char *)dp + index;
}
		
char *String::getText(void) const
{
	if(isBig())
		return (char *)content.bigstring.text;

	return (char *)content.ministring.text;
}

const unsigned String::getSize(void) const
{
	if(isBig())
		return content.bigstring.size;

	return minsize;
}

void String::setLength(unsigned len)
{
	if(isBig())
		content.bigstring.length = len;
	else
		content.ministring.length = len;
}

unsigned String::setSize(unsigned size)
{
	if(size <= minsize && !isBig())
		return minsize;

	if(size <= slotlimit)
	{
		size = size / slotsize;
		++size *= slotsize;
	}
	content.bigstring.size = size;
	return size;
}

const unsigned String::getLength(void) const
{
	if(isBig())
		return content.bigstring.length;

	return content.ministring.length;
}

String operator+(const String &s1, char c2)
{
	String result(s1);
	result.add(c2);
	return result;
}

String operator+(const String &s1, const String &s2)
{
	String result(s1);
	result.append(s2);
	return result;
}

String operator+(const String &s1, const char *s2)
{
	String result(s1);
	result += s2;
	return result;
}

String operator+(const char *s1, const String &s2)
{
	String result(s1);
	result += s2;
	return result;
}

String operator+(char c1, const String &s2)
{
	String result(1, c1);
	result += s2;
	return result;
}

bool String::operator<(const String &s1)
{
        if(compare(s1.getText()) < 0)
                return true;

        return false;
}

bool String::operator<(const char *s1)
{
        if(compare(s1) < 0)
                return true;

        return false;
}

bool String::operator>(const String &s1)
{
        if(compare(s1.getText()) > 0)
                return true;

        return false;
}

bool String::operator>(const char *s1)
{
        if(compare(s1) > 0)
                return true;

        return false;
}

bool String::operator<=(const String &s1)
{
	if(compare(s1.getText()) <= 0)
		return true;

	return false;
}

bool String::operator<=(const char *s1)
{
        if(compare(s1) <= 0)
                return true;

        return false;
}

bool String::operator>=(const String &s1)
{
        if(compare(s1.getText()) >= 0)
                return true;

        return false;
}

bool String::operator>=(const char *s1)
{
        if(compare(s1) >= 0)
                return true;

        return false;
}

bool String::operator==(const String &s1)
{
        if(compare(s1.getText()) == 0)
                return true;

        return false;
}

bool String::operator==(const char *s1)
{
        if(compare(s1) == 0)
                return true;

        return false;
}

bool String::operator!=(const String &s1)
{
        if(compare(s1.getText()) != 0)
                return true;

        return false;
}

bool String::operator!=(const char *s1)
{
        if(compare(s1) != 0)
                return true;

        return false;
}

bool String::operator*=(const String &s1)
{
	return search(s1.getText(), s1.getLength()) != npos;
}

bool String::operator*=(const char *s)
{
	return search(s) != npos;
}

std::ostream &operator<<(std::ostream &os, const String &str)
{
	os << str.getText();
	return os;
}

std::istream &getline(std::istream &is, String &str, char delim, unsigned len)
{
	if(!len)
		len = str.getSize() - 1;

	if(len >= str.getSize())
		str.resize(len + 1);

	char *ptr = str.getText();
	is.getline(ptr, len, delim);
	str.setLength(strlen(ptr));
	return is;
}

SString::SString() :
String(), streambuf()
#ifdef  HAVE_OLD_IOSTREAM
,ostream()
#else
,ostream((streambuf *)this)
#endif
{
#ifdef  HAVE_OLD_IOSTREAM
        ostream::init((streambuf *)this);
#endif
}

SString::SString(const SString &from) :
String(from), streambuf()
#ifdef  HAVE_OLD_IOSTREAM
,ostream()
#else
,ostream((streambuf *)this)
#endif
{
#ifdef  HAVE_OLD_IOSTREAM
        ostream::init((streambuf *)this);
#endif
}

SString::~SString()
{
	if(isBig())
		String::clear();
}

int SString::overflow(int c)
{
	String::add((char)(c));
	return c;
}

#ifdef	HAVE_SNPRINTF
int strprintf(String &str, unsigned size, const char *format, ...)
{
	va_list args;
	va_start(args, format);
	int rtn;

	if(!size)
		size = str.getSize();

	if(size > str.getSize())
		str.resize(size);

	char *ptr = str.getText();
	str.setLength(0);
	ptr[0] = 0;
	rtn = vsnprintf(ptr, size, format, args);
	str.setLength(strlen(ptr));
	va_end(args);
	return rtn;
}
#endif

#ifdef	CCXX_NAMESPACES
};	
#endif

/** EMACS **
 * Local variables:
 * mode: c++
 * c-basic-offset: 8
 * End:
 */
